from PyQt5 import QtOpenGL, QtCore, QtGui
from OpenGL.GL import *
from OpenGL.GLU import *
from geometry.pnt3d import Pnt3D
from t3d import T3d
import math


class GLCanvas(QtOpenGL.QGLWidget):

    def __init__(self, _controller=[], _view=[]):
        super(GLCanvas, self).__init__()

        # canvas sizes
        self.width = 0   # width of canvas (horizontal raster size)
        self.height = 0  # height of canvas (vertical raster size)

        # camera (eye) parameters
        self.eyex = 0.0    # camera (eye) x coordinate in object space
        self.eyey = 0.0    # camera (eye) y coordinate in object space
        self.eyez = 1.0    # camera (eye) z coordinate in object space
        self.refx = 0.0    # reference point x coordinate in object space
        self.refy = 0.0    # reference point y coordinate in object space
        self.refz = 0.0    # reference point z coordinate in object space
        self.vupx = 0.0    # view up-vector x component in object space
        self.vupy = 0.0    # view up-vector y component in object space
        self.vupz = 1.0    # view up-vector z component in object space

        # model clipping volume limits
        self.left = -1.0    # left limit of clipping volume (eye space)
        self.right = 1.0    # right limit of clipping volume (eye space)
        self.bottom = -1.0  # bottom limit of clipping volume (eye space)
        self.top = 1.0      # top limit of clipping volume (eye space)
        self.front = 1.0    # front limit of clipping volume (eye space)
        self.back = 2.0     # back limit of clipping volume (eye space)

        # Set lighting flag
        self.lighting = True

        # Set perspective projection flag
        self.perspective = True

        # Set cull face flag
        self.cullFace = False

        # Set depth test flag
        self.depthTest = True

        # 3D transformation object
        self.t3d = T3d()

        # Eye coordinate system unit vector
        self.xe = Pnt3D()
        self.ye = Pnt3D()
        self.ze = Pnt3D()

        # mouse related properties
        # (Mouse buttons: QCore.Qt.LeftButton, QCore.Qt.RightButton, and
        #  QCore.Qt.MidButton)
        self.mouseButton = QtCore.Qt.NoButton
        self.mousebuttonPressed = False  # if true, mouse button is pressed
        self.pt0 = QtCore.QPoint(0, 0)  # mouse position at button press event
        self.pt1 = QtCore.QPoint(0, 0)  # current mouse position
        self.prevPt = QtCore.QPoint(0, 0)  # previous mouse move position
        self.mouseMoveTol = 2  # tolerance for mouse move

        # factor for zooming (in or out) clipping window (in eye coordinate
        # system) with respect to current window sizes
        self.zoomFac = 0.1

        # Mouse pan move flag
        self.panMoveOn = False  # flag for mouse pan move

        # Pressed key properties
        self.shiftKeyPressed = False

        # canvas aggregation objects
        self.controller = _controller  # handle to app controller
        self.view = _view  # handle to app view

        # OpenGL display lists and display update flags
        self.viewDsp = 0  # GL list index for model display
        self.updatedDsp = False  # if true, model display is updated

    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # --------------------GENERAL PROPERTY METHODS-------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def setController(self, _controller):
        self.controller = _controller

    # ---------------------------------------------------------------------
    def getController(self):
        return self.controller

    # ---------------------------------------------------------------------
    def setView(self, _view):
        self.view = _view

    # ---------------------------------------------------------------------
    def getView(self):
        return self.view

    # ---------------------------------------------------------------------
    def resetViewDisplay(self):
        if self.view is None:
            return
        self.updatedDsp = False
        self.update()

    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # --------------------CANVAS PREDEFINED SLOTS--------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def initializeGL(self):
        # Set lighting and depth test parameters
        if self.lighting:
        glEnable(GL_LIGHTING)
        else:
            glDisable(GL_LIGHTING)
        glEnable(GL_LIGHT0)
        glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE)
        glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE)
        glEnable(GL_COLOR_MATERIAL)
        if self.cullFace:
            glEnable(GL_CULL_FACE)
        else:
            glDisable(GL_CULL_FACE)
        if self.depthTest:
            glEnable(GL_DEPTH_TEST)
        else:
            glDisable(GL_DEPTH_TEST)

        # set canvas background color
        color = self.view.getBackgroundColor()
        glClearColor(color[0], color[1], color[2], 1.0)
        glClear(GL_COLOR_BUFFER_BIT)

    # ---------------------------------------------------------------------
    def resizeGL(self, _width, _height):

        # Avoid division by zero
        if _width == 0:
            _width = 1

        # store GL canvas sizes in object properties
        self.width = _width
        self.height = _height

        # Setup visualization parameters based on model bounding box
        if (self.view is not None) and (not self.view.isEmpty()):
            self.resetView()

        # Setup the viewport to canvas dimesions
        glViewport(0, 0, self.width, self.height)

        # Setup (clipping) view volume and type of projection (orthotropic or perpective)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        if self.perspective:
            glFrustum(self.left, self.right, self.bottom, self.top, self.front, self.back)
        else:
            glOrtho(self.left, self.right, self.bottom, self.top, self.front, self.back)


    # ---------------------------------------------------------------------
    def paintGL(self):

        # clear the buffer with the current color
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        # Check to see whether there is a model
        if (self.view is None) or self.view.isEmpty():
            return

        # Setup default head light position (before applying camara
        # transformation) and setup camera parameters
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        light_pos = [0.0, 0.0, 1.0, 0.0]
        glLightfv( GL_LIGHT0, GL_POSITION, light_pos )
        gluLookAt( self.eyex, self.eyey, self.eyez, \
                   self.refx, self.refy, self.refz, \
                   self.vupx, self.vupy, self.vupz )

        # Display view , storing it in a display list
        if not self.updatedDsp:
            if self.viewDsp > 0:
                glDeleteLists(self.viewDsp, 1)
            self.viewDsp = self.makeDisplayView()

        if self.viewDsp > 0:
            glCallList(self.viewDsp)
            self.updatedDsp = True

    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # -----------------------DISPLAY FUNCTIONS-----------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def drawFaces(self):
        faces = self.view.getFaces()
        for i in range(0, len(faces)):
            fcolors = self.view.getFaceColors(i)
            fnormals = self.view.getFaceNormals(i)
            pts = self.view.getFacePts(i)
            if len(pts) == 3:
                glBegin(GL_TRIANGLES)
            elif len(pts) == 4:
                glBegin(GL_QUADS)
            else:
                glBegin(GL_POLYGON)
            for j in range(len(pts)):
                glColor3f(fcolors[j][0], fcolors[j][1], fcolors[j][2])
                glNormal3f(fnormals[j][0], fnormals[j][1], fnormals[j][2])
                glVertex3f(pts[j].getX(), pts[j].getY(), pts[j].getZ())
            glEnd()

    # ---------------------------------------------------------------------
    def makeDisplayView(self):
        if (self.view is None) or (self.view.isEmpty()):
            return 0

        list = glGenLists(1)
        glNewList(list, GL_COMPILE)
        self.drawFaces()
        glEndList()
        return list

    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ------------ METHODS TO MANAGE VISUALIZATION PARAMETERS -------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def resetView(self):
        if (self.view is None) or self.view.isEmpty():
            return

        # Get model bounding box
        xmin, xmax, ymin, ymax, zmin, zmax = self.view.getBoundBox()

        # Get model max box size
        sizex = xmax - xmin
        sizey = ymax - ymin
        sizez = zmax - zmin
        if sizex > sizey:
            max_size = sizex
        else:
            max_size = sizey
        if max_size < sizez:
            max_size = sizez

        # Place reference point at center of bounding box.
        # **** COMPLETE HERE - GLCANVAS: 01 ****

        # Place camera with same x and y coordinates of reference point,
        # distant 4 "max_sizes" to reference point in z direction
        # **** COMPLETE HERE - GLCANVAS: 02 ****

        #  Define camera vertical plane perpendicular to xz plane.
        # **** COMPLETE HERE - GLCANVAS: 03 ****

        # Compute canvas viewport aspect ratio.
        vpr = float(self.height) / float(self.width)

        # Define view window with minimum size equal to model
        # maximum size and with aspect ratio equal to canvas.
        view_window_w = view_window_h = max_size
        # **** COMPLETE HERE - GLCANVAS: 04 ****

        # Define view volume parameters such that:
        # (1) view window center coincides with canvas center
        # (2) front clip plane is located at 3 "max_sizes" from camera
        # (3) back  clip plane is located at 5 "max_sizes" from camera
        # **** COMPLETE HERE - GLCANVAS: 05 ****

        # Build eye coordinate unit vectors
        self.buildEyeSystem()

    # ---------------------------------------------------------------------
    def buildEyeSystem(self):
        # view vector and view upper vector:
        view = Pnt3D()
        upvec = Pnt3D()

        # Define view vector (from ref to eye) and view up vector.
        # **** COMPLETE HERE - GLCANVAS: 06 ****

        # Find the eye system unit vectors.
        # **** COMPLETE HERE - GLCANVAS: 07 ****

    # ---------------------------------------------------------------------
    def scaleEyeWindow(self, _scaleFac):
        # Get current window center.
        # **** COMPLETE HERE: GLCANVAS_08 ****

        # Set new window sizes based on scaling factor
        # **** COMPLETE HERE: GLCANVAS_09 ****

        # Adjust window limits
        # **** COMPLETE HERE: GLCANVAS_10 ****

        # Setup (clipping) view volume in eye coordinate system
        # and type of projection (orthotropic or perpective)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        if self.perspective:
            glFrustum(self.left, self.right, self.bottom, self.top, self.front, self.back)
        else:
            glOrtho(self.left, self.right, self.bottom, self.top, self.front, self.back)

        # Force redisplay
        self.update()

    # ---------------------------------------------------------------------
    def panEyeWindow(self, _panX, _panY):
        # Shift current window.
        # **** COMPLETE HERE: GLCANVAS_11 ****

        # Setup (clipping) view volume in eye coordinate system
        # and type of projection (orthotropic or perpective)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        if self.perspective:
            glFrustum(self.left, self.right, self.bottom, self.top, self.front, self.back)
        else:
            glOrtho(self.left, self.right, self.bottom, self.top, self.front, self.back)

        # Force redisplay
        self.update()

    # ---------------------------------------------------------------------
    def buildWorlToEyeTransform(self):
        # Create a transformation matrix for current camera transformation.
        self.t3d.identity()
        # **** COMPLETE HERE: GLCANVAS_12 ****
        if self.perspective:
            # **** COMPLETE HERE: GLCANVAS_13 ****

    # ---------------------------------------------------------------------
    def rotateViewAboutX(self, _mouseMoveFacY):
        # Estimate rotation angle based on mouse move factor
        angle = _mouseMoveFacY * 180.0

        # Rotate eye point around eye x axis about reference point.
        self.t3d.identity()
        # **** COMPLETE HERE: GLCANVAS_14 ****
        self.eyex, self.eyey, self.eyez = self.t3d.transform(self.eyex, self.eyey, self.eyez)

        # Rotate view up vector around eye x axis.
        self.t3d.identity()
        # **** COMPLETE HERE: GLCANVAS_15 ****
        self.vupx, self.vupy, self.vupz = self.t3d.transform(self.vupx, self.vupy, self.vupz)

        # Rebuild eye coordinate unit vectors
        self.buildEyeSystem()

        # Force redisplay
        self.update()

    # ---------------------------------------------------------------------
    def rotateViewAboutY(self, _mouseMoveFacX):
        # Estimate rotation angle based on mouse move factor
        angle = _mouseMoveFacX * 180.0

        # Rotate eye point around eye y axis about reference point.
        self.t3d.identity()
        # **** COMPLETE HERE: GLCANVAS_16 ****
        self.eyex, self.eyey, self.eyez = self.t3d.transform(self.eyex, self.eyey, self.eyez)

        # Rotate view up vector around eye y axis.
        self.t3d.identity()
        # **** COMPLETE HERE: GLCANVAS_17 ****
        self.vupx, self.vupy, self.vupz = self.t3d.transform(self.vupx, self.vupy, self.vupz)

        # Rebuild eye coordinate unit vectors
        self.buildEyeSystem()

        # Force redisplay
        self.update()

    # ---------------------------------------------------------------------
    def rotateViewAboutZ(self, _prevPt, _currPt):
        # Transform reference point from world to eye system.
        self.buildWorlToEyeTransform()
        refxInEye, refyInEye, refzInEye = self.t3d.transform(self.refx, self.refy, self.refz)

        # Compute the spin angle based on the increment of mouse
        # position about reference point projected to eye window.
        angle = math.atan2(_currPt.y() - refyInEye, _currPt.x() - refxInEye) - \
                math.atan2(_prevPt.y() - refyInEye, _prevPt.x() - refxInEye)

        # Transform angle to degree and invert sign.
        angle = -angle * (180.0 / math.pi)

        # Rotate view up vector around eye z axis.
        self.t3d.identity()
        # **** COMPLETE HERE: GLCANVAS_18 ****
        self.vupx, self.vupy, self.vupz = self.t3d.transform(self.vupx, self.vupy, self.vupz)

        # Rebuild eye coordinate unit vectors
        self.buildEyeSystem()

        # Force redisplay
        self.update()

    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # --------FUNCTION TO CONVERT POINT FROM RASTER TO WORLD COORDS--------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def convertRasterPtToEyeWindow(self, _pt):
        dX = self.right - self.left  # World window horizontal size
        dY = self.top - self.bottom  # World window vertical size

        # Origin of canvas raster coordinates it at the left-top corner,
        # while origin of GL canvas floating point coordinates is at
        # left-bottom corner.
        # mX is the distance of point to left universe window limit
        # mY is the distance of point to bottom universe window limit
        mX = float(_pt.x()) * dX / float(self.width)
        my = float(self.height - _pt.y()) * dY / float(self.height)
        x = self.left + mX
        y = self.bottom + my
        return QtCore.QPointF(x, y)
 
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ----------------------MOUSE EVENT SLOTS------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def mousePressEvent(self, event):
        self.mouseButton = event.button()
        self.mousebuttonPressed = True

        # Get current mouse position
        self.pt0 = event.pos()
        self.pt1 = self.pt0

        # Verify whether the middle button is turned on.
        # If this is the case, change mouse cursor feedback, save current
        # mouse position for computing pan translation in next mouse-move,
        # set panMove flag as true, and continue to next mouse-move event.
        if self.mouseButton == QtCore.Qt.MiddleButton:
            self.setCursor(QtGui.QCursor(QtCore.Qt.SizeAllCursor))
            self.prevPt = self.pt0
            self.panMoveOn = True
            return

    # ---------------------------------------------------------------------
    def mouseMoveEvent(self, event):
        # Disregard current point if not button pressed
        if not self.mousebuttonPressed:
            return

        # Save current mouse position
        self.prevPt = self.pt1

        # Get current mouse position
        self.pt1 = event.pos()

        # Disregard current point if current point is not at the same
        # location of button press point
        if((abs(self.prevPt.x() - self.pt1.x()) <= self.mouseMoveTol) and
           (abs(self.prevPt.y() - self.pt1.y()) <= self.mouseMoveTol)):
           return

        # Compute mouse move in horizontal and vertical directions
        # in eye coordinate system
        ptPrevEye = self.convertRasterPtToEyeWindow(self.prevPt)
        pt1Eye = self.convertRasterPtToEyeWindow(self.pt1)

        # Verify whether panMoveOn is true and process panMouseMove
        if self.panMoveOn:
            # Shift clipping window in horizontal and vertical directions
            # based on mouse position move
            panX = pt1Eye.x() - ptPrevEye.x()
            panY = pt1Eye.y() - ptPrevEye.y()

            # In case shift key is pressed, disregard smaller mouse
            # move component
            if self.shiftKeyPressed:
                if abs(panX) > abs(panY):
                    panY = 0.0
                else:
                    panX = 0.0

            # Pan eye window in opposite direction of mouse move
            self.panEyeWindow(-panX, -panY)
            return

        # If left mouse button is pressed, rotate view around horizontal
        # or vertical screen axes about reference point.
        if self.mouseButton == QtCore.Qt.LeftButton:
            # Compute normalized mouse translation and decide to rotate about
            # horizontal or vertical screen axes based on greatest mouse translation
            mouseMoveFacX = float(self.pt1.x() - self.prevPt.x()) / float(self.width)
            mouseMoveFacY = float(self.pt1.y() - self.prevPt.y()) / float(self.height)
            if abs(mouseMoveFacX) > abs(mouseMoveFacY):
                self.rotateViewAboutY(-mouseMoveFacX)
            else:
                self.rotateViewAboutX(-mouseMoveFacY)

        # If right mouse button is pressed, rotate view around out of plane
        # screen axis about reference point.
        elif self.mouseButton == QtCore.Qt.RightButton:
            self.rotateViewAboutZ(ptPrevEye, pt1Eye)

    # ---------------------------------------------------------------------
    def mouseReleaseEvent(self, event):
        self.mousebuttonPressed = False
        # Finish panMove
        if self.panMoveOn:
            self.panMoveOn = False
            self.setCursor(QtGui.QCursor(QtCore.Qt.ArrowCursor))
            return

    # ---------------------------------------------------------------------
    def wheelEvent(self, event):
        if event.angleDelta().y() > 0:
            scaleFac = 1.0 - self.zoomFac
        else:
            scaleFac = 1.0 + self.zoomFac
        self.scaleEyeWindow(scaleFac)

    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ----------------------KEYBOARD EVENT SLOTS---------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------
    # ---------------------------------------------------------------------

    # ---------------------------------------------------------------------
    def keyPressEvent(self, event):
        self.shiftKeyPressed = (event.key() == QtCore.Qt.Key_Shift)

    # ---------------------------------------------------------------------
    def keyReleaseEvent(self, event):

        # If Ctrl-H key was pressed, pop-up help message.
        if event.key() == QtCore.Qt.Key_H:
            textOpenFile = 'Ctrl-O ==> Open fac file\n'
            if self.lighting:
                textLighting = 'Ctrl-L ==> Disable lighting\n'
            else:
                textLighting = 'Ctrl-L ==> Enable lighting\n'
            if self.perspective:
                textPerspective = 'Ctrl-P ==> Disable perspective\n'
            else:
                textPerspective = 'Ctrl-P ==> Enable perspective\n'
            if self.cullFace:
                textCullFace = 'Ctrl-C ==> Disable cull face\n'
            else:
                textCullFace = 'Ctrl-C ==> Enable cull face\n'
            if self.depthTest:
                textDepthTest = 'Ctrl-D ==> Disable depth test'
            else:
                textDepthTest = 'Ctrl-D ==> Enable depth test'
            helpMessage = textOpenFile + textLighting + textPerspective + textCullFace + textDepthTest
            self.controller.popupMessage(helpMessage)
            return

        # If Ctrl-O key was pressed, open fac file.
        if event.key() == QtCore.Qt.Key_O:
            self.controller.openFile()
            self.resetView()
            glMatrixMode(GL_PROJECTION)
            glLoadIdentity()
            if self.perspective:
                glFrustum(self.left, self.right, self.bottom, self.top, self.front, self.back)
            else:
                glOrtho(self.left, self.right, self.bottom, self.top, self.front, self.back)
            self.resetViewDisplay()
            self.update()
            return

        # If Ctrl-L key was pressed, toggle (on/off) lighting.
        if event.key() == QtCore.Qt.Key_L:
            if self.lighting:
                self.lighting = False
                glDisable(GL_LIGHTING)
            else:
                self.lighting = True
                glEnable(GL_LIGHTING)
            self.update()
            return

        # If Ctrl-P key was pressed, toggle perspective/orthografic projection.
        if event.key() == QtCore.Qt.Key_P:
            if self.perspective:
                self.perspective = False
                glMatrixMode(GL_PROJECTION)
                glLoadIdentity()
                glOrtho(self.left, self.right, self.bottom, self.top, self.front, self.back)
            else:
                self.perspective = True
                glMatrixMode(GL_PROJECTION)
                glLoadIdentity()
                glFrustum(self.left, self.right, self.bottom, self.top, self.front, self.back)
            self.update()
            return

        # If Ctrl-C key was pressed, toggle enable/disable cull face.
        if event.key() == QtCore.Qt.Key_C:
            if self.cullFace:
                self.cullFace = False
                glDisable(GL_CULL_FACE)
            else:
                self.cullFace = True
                glEnable(GL_CULL_FACE)
            self.update()
            return
 
        # If Ctrl-D key was pressed, toggle enable/disable depth test.
        if event.key() == QtCore.Qt.Key_D:
            if self.depthTest:
                self.depthTest = False
                glDisable(GL_DEPTH_TEST)
            else:
                self.depthTest = True
                glEnable(GL_DEPTH_TEST)
            self.update()
            return

        # If only shift key was pressed, setup flag
        self.shiftKeyPressed = False
