# ----------------------------------------------------------------------------
# t3d.py  -  3D matrix transformations class.
# ----------------------------------------------------------------------------
#
# ----------------------------------------------------------------------------
# Reference:
#
# See class notes of courses CIV-2802 Sistemas Graficos para
# Engenharia, Departamento de Engenharia Civil, PUC-Rio.
#
# ----------------------------------------------------------------------------
# Functionality:
#
# This package maintains a 3D transformation 4x4 matrix which can
# be initialized as the identity matrix, and which accumulates 
# the usual 3D geometric transformations and 3D visualization 
# transformations.
#
# There is a public function to transform a given 3D point 
# according to the current transformation matrix.
#
# The 3D transformations considered are scaling, translation, 
# rotation of basis, camera transformation, perspective 
# projection, and view volume normalization.  
# There is a public function for each one of these transformations.  
# The effect of these functions is to accumulate the corresponding 
# transformation in the current transformation matrix.
#
# ----------------------------------------------------------------------------

from geometry.pnt3d import Pnt3D
import numpy as np
import math


class T3d:
    def __init__(self):
        self.mt = np.zeros((4,4), dtype='float')
        self.identity()

# ----------------------------------------------------------------------------
# This method sets the current transformation matrix as an identity matrix.
#
    def identity(self):
        self.mt[0][0] = 1.0; self.mt[0][1] = 0.0; self.mt[0][2] = 0.0; self.mt[0][3] = 0.0
        self.mt[1][0] = 0.0; self.mt[1][1] = 1.0; self.mt[1][2] = 0.0; self.mt[1][3] = 0.0
        self.mt[2][0] = 0.0; self.mt[2][1] = 0.0; self.mt[2][2] = 1.0; self.mt[2][3] = 0.0
        self.mt[3][0] = 0.0; self.mt[3][1] = 0.0; self.mt[3][2] = 0.0; self.mt[3][3] = 1.0

# ----------------------------------------------------------------------------
# This function pre-multiplies the given matrix to the current transformation
# matrix.
#
    def accumulate(self, new_mt):
        t0            =  self.mt[0][0] * new_mt[0][0] + self.mt[1][0] * new_mt[0][1] \
                       + self.mt[2][0] * new_mt[0][2] + self.mt[3][0] * new_mt[0][3]
        t1            =  self.mt[0][0] * new_mt[1][0] + self.mt[1][0] * new_mt[1][1] \
                       + self.mt[2][0] * new_mt[1][2] + self.mt[3][0] * new_mt[1][3]
        t2            =  self.mt[0][0] * new_mt[2][0] + self.mt[1][0] * new_mt[2][1] \
                       + self.mt[2][0] * new_mt[2][2] + self.mt[3][0] * new_mt[2][3]
        self.mt[3][0] =  self.mt[0][0] * new_mt[3][0] + self.mt[1][0] * new_mt[3][1] \
                       + self.mt[2][0] * new_mt[3][2] + self.mt[3][0] * new_mt[3][3]
        self.mt[2][0] =  t2
        self.mt[1][0] =  t1
        self.mt[0][0] =  t0

        t0            =  self.mt[0][1] * new_mt[0][0] + self.mt[1][1] * new_mt[0][1] \
                       + self.mt[2][1] * new_mt[0][2] + self.mt[3][1] * new_mt[0][3]
        t1            =  self.mt[0][1] * new_mt[1][0] + self.mt[1][1] * new_mt[1][1] \
                       + self.mt[2][1] * new_mt[1][2] + self.mt[3][1] * new_mt[1][3]
        t2            =  self.mt[0][1] * new_mt[2][0] + self.mt[1][1] * new_mt[2][1] \
                       + self.mt[2][1] * new_mt[2][2] + self.mt[3][1] * new_mt[2][3]
        self.mt[3][1] =  self.mt[0][1] * new_mt[3][0] + self.mt[1][1] * new_mt[3][1] \
                       + self.mt[2][1] * new_mt[3][2] + self.mt[3][1] * new_mt[3][3]
        self.mt[2][1] =  t2
        self.mt[1][1] =  t1
        self.mt[0][1] =  t0

        t0            =  self.mt[0][2] * new_mt[0][0] + self.mt[1][2] * new_mt[0][1] \
                       + self.mt[2][2] * new_mt[0][2] + self.mt[3][2] * new_mt[0][3]
        t1            =  self.mt[0][2] * new_mt[1][0] + self.mt[1][2] * new_mt[1][1] \
                       + self.mt[2][2] * new_mt[1][2] + self.mt[3][2] * new_mt[1][3]
        t2            =  self.mt[0][2] * new_mt[2][0] + self.mt[1][2] * new_mt[2][1] \
                       + self.mt[2][2] * new_mt[2][2] + self.mt[3][2] * new_mt[2][3]
        self.mt[3][2] =  self.mt[0][2] * new_mt[3][0] + self.mt[1][2] * new_mt[3][1] \
                       + self.mt[2][2] * new_mt[3][2] + self.mt[3][2] * new_mt[3][3]
        self.mt[2][2] =  t2
        self.mt[1][2] =  t1
        self.mt[0][2] =  t0

        t0            =  self.mt[0][3] * new_mt[0][0] + self.mt[1][3] * new_mt[0][1] \
                       + self.mt[2][3] * new_mt[0][2] + self.mt[3][3] * new_mt[0][3]
        t1            =  self.mt[0][3] * new_mt[1][0] + self.mt[1][3] * new_mt[1][1] \
                       + self.mt[2][3] * new_mt[1][2] + self.mt[3][3] * new_mt[1][3]
        t2            =  self.mt[0][3] * new_mt[2][0] + self.mt[1][3] * new_mt[2][1] \
                       + self.mt[2][3] * new_mt[2][2] + self.mt[3][3] * new_mt[2][3]
        self.mt[3][3] =  self.mt[0][3] * new_mt[3][0] + self.mt[1][3] * new_mt[3][1] \
                       + self.mt[2][3] * new_mt[3][2] + self.mt[3][3] * new_mt[3][3]
        self.mt[2][3] =  t2
        self.mt[1][3] =  t1
        self.mt[0][3] =  t0

# ----------------------------------------------------------------------------
# This method transforms the given coordinates according to the current
# transformation matrix.
# The result of the transformation is divided by the fourth homogeneous
# coordinate.
#
    def transform(self, _x, _y, _z ):
        r = self.mt[0][0] * _x + self.mt[0][1] * _y + self.mt[0][2] * _z + self.mt[0][3]
        s = self.mt[1][0] * _x + self.mt[1][1] * _y + self.mt[1][2] * _z + self.mt[1][3]
        t = self.mt[2][0] * _x + self.mt[2][1] * _y + self.mt[2][2] * _z + self.mt[2][3]
        h = self.mt[3][0] * _x + self.mt[3][1] * _y + self.mt[3][2] * _z + self.mt[3][3]
        x = r / h
        y = s / h
        z = t / h
        return x, y, z

# ----------------------------------------------------------------------------
# This method accumulates a scaling transformation to the current
# transformation matrix.
#
    def scale(self, sx, sy, sz):
        scale = np.zeros((4,4), dtype='float')

        # **** COMPLETE HERE - T3D: 01 ****

        self.accumulate(scale)

# ----------------------------------------------------------------------------
# This method accumulates a translation transformation to the current
# transformation matrix.
#
    def translate(self, tx, ty, tz):
        trans = np.zeros((4,4), dtype='float')

        # **** COMPLETE HERE - T3D: 02 ****

        self.accumulate(trans)

# ----------------------------------------------------------------------------
# This method accumulates a rotation transformation to the current
# transformation matrix.  The rotation is performed around a given vector
# through the origin.  The vector direction is indicated by its coordinates.
# The rotation angle in assumed given in degrees.
#    @param
#    angle      - rotation angle (degrees)                     (in)
#    vx, vy, vz - vector direction coordinates                 (in)
#
    def rotateVecOrg(self, angle, vx, vy, vz):
        cost = math.cos(math.radians(angle))
        sint = math.sin(math.radians(angle))
        v = Pnt3D(vx, vy, vz)
        v = Pnt3D.normalize(v)
        rot = np.zeros((4,4), dtype='float')
 
        # **** COMPLETE HERE - T3D: 03 ****

        self.accumulate(rot)

# ----------------------------------------------------------------------------
# This method accumulates a rotation transformation to the current 
# transformation matrix.  The rotation is performed around a given vector
# through a given point.  The vector direction and the point of reference for
# rotation are indicated by their coordinates.
# The rotation angle in assumed given in degrees.
#    @param
#    angle      - rotation angle (degrees)                     (in)
#    vx, vy, vz - vector direction coordinates                 (in)
#    px, py, pz - point coordinates                            (in)
#
    def rotateVecPnt(self, angle, vx, vy, vz, px, py, pz):
        # Translate to origin.
        # **** COMPLETE HERE - T3D: 04 ****

        # Rotate about given vector direction.
        # **** COMPLETE HERE - T3D: 05 ****

        # Translate back to original position.
        # **** COMPLETE HERE - T3D: 06 ****

# ----------------------------------------------------------------------------
# This method accumulates a rotation of basis transformation to the current
# transformation matrix.
# The rotation of basis is a rotation transformation whose matrix is such
# that each row of the 3x3 linear transformation submatrix is formed by the
#  iven new basis unitary vector coordinates.
#
    def rotateBasis(self, ux, uy, uz, vx, vy, vz, wx, wy, wz):
        rot = np.zeros((4,4), dtype='float')

        # **** COMPLETE HERE - T3D: 07 ****

        self.accumulate(rot)

# ----------------------------------------------------------------------------
# This method accumulates a camara transformation to the current
# transformation matrix.  The camara transformation is a change-of-basis
# transformation formed by a concatenation of a translation from the origin
# of the world (client) coordinate sytem to the eye point and a rotation
# from the world coordinatesystem to the eye coordinate system (a
# right-handed system  centered at the eye point, with the z axis from the
# reference point to the eye point).
# The given view upper vector defines with the view direction  (eye system
# z axis, going from the reference point to the eye point) the eye system
# yz plane.
#    @param
#    eyex, eyey, eyez     - eye position                     (in)
#    @param
#    refx, refy, refz     - reference point position         (in)
#    @param
#    vupx, vupy, vupz     - view up vector direction         (in)
#
    def camera(self, eyex, eyey, eyez, refx, refy, refz, vupx, vupy, vupz):
        # view vector and view upper vector:
        view = Pnt3D()
        upvec = Pnt3D()
        # eye system unit vectors:
        xe = Pnt3D()
        ye = Pnt3D()
        ze = Pnt3D()

        # Define view vector (from ref to eye) and view up vector.
        # **** COMPLETE HERE - T3D: 08 ****

        # Find the eye system unit vectors.
        # **** COMPLETE HERE - T3D: 09 ****

        # Concatenate a translation of basis from origin to eye point.
        # **** COMPLETE HERE - T3D: 10 ****

        # Concatenate a rotation of basis from world to eye system.
        # **** COMPLETE HERE - T3D: 11 ****

# ----------------------------------------------------------------------------
# This method accumulates a canonical perspective transformation to the
# current transformation matrix.  This transformation defines a conical
# projection whose center of projection is assumed as the origin of the
# system of coordinates.  Alternatively, this transformation defines a 
# distortion of the coordinate space such that the result of a orthographic
# projection of the point in the distorted space coincides with the result
# of the conical projection.
# The plane of projection and the two frustum clipping planes (front and back)
# are assumed normal to the z axis, lying on its negative side.  These planes
# are defined by their distances to the origin.  
# It is assumed that both values for the clipping plane positions are positive
# and that the visible part of the 3D space lies along the negative side of
# the z-axis in the eye coordinate system.
# The perspective distortion is such that it does not change z coordinate
# values of points lying on the two frustum clipping planes.
#    @param
#    d           -  distance of projection plane to origin   (in)
#    @param
#    front       -  frustum front clipping plane position    (in)
#    @param
#    back        -  frustum back clipping plane position     (in)
#
    def perspective(self, d, front, back):
        pers = np.zeros((4,4), dtype='float')

        # Define alpha and beta perspective parameters from the
        # front and back arguments.
        # **** COMPLETE HERE - T3D: 12 ****

        # Create Perspective matrix.
        # **** COMPLETE HERE - T3D: 13 ****

        self.accumulate(pers)

# ----------------------------------------------------------------------------
# This method accumulates a view volume normalization transformation to the
# current transformation matrix.  This normalization is such that it
# transforms a view volume defined in the eye (camera) coordinate system
# into a cube whose center is at the origin and whose coordinates are in
# the range from -1 to +1.
# The given view volume defines, in the eye coordinate system, the portion
# of modeling space which is seen on the screen. That is, all points lying
# ouside the view volume are clipped.
# It is assumed that both values for the deepness clipping plane positions
# are positive and that the visible part of the 3D space lies along the
# negative side of the z-axis in the eye coordinate system.
# The normalization transformation is such that points at the front
# clipping plane will have z-coordinates equal to -1 and points at the
# back clipping plane will have z-coordinates equal to +1.  Therefore,
# there is a reflection transformation with respect to the xy plane in the 
# view volume normalization transformation.
#    @param
#    left, right - horizontal clipping limits of view volume (in)
#    @param
#    bottom, top - vertical clipping limits of view volume   (in)
#    @param
#    front, back - deepness clipping limits of view volume   (in)
#
    def normalizeView(self, left, right, bottom, top, front, back):

        # Concatenate a translation of basis from origin to center
        # of view volume.
        # **** COMPLETE HERE - T3D: 14 ****

        # Concatenate a scaling transformation that scales the view 
        # volume to a cube ranging from -1 to +1 in the 3 directions, 
        # also reverting the z-axis direction.
        # **** COMPLETE HERE - T3D: 15 ****
